/*
 * Project: Measuring Relative Humidity and Temperature using Sensirion's
   SHT11 and SHT75 sensors
   Description:
   The two sensors are connected to PIC18F2550 microcontroller.
   PIC18F2550 runs at 48.0 MHz

   Date: May 24, 2011

 */

// LCD module connections
sbit LCD_RS at RC6_bit;
sbit LCD_EN at RC7_bit;
sbit LCD_D4 at RB4_bit;
sbit LCD_D5 at RB5_bit;
sbit LCD_D6 at RB6_bit;
sbit LCD_D7 at RB7_bit;

sbit LCD_RS_Direction at TRISC6_bit;
sbit LCD_EN_Direction at TRISC7_bit;
sbit LCD_D4_Direction at TRISB4_bit;
sbit LCD_D5_Direction at TRISB5_bit;
sbit LCD_D6_Direction at TRISB6_bit;
sbit LCD_D7_Direction at TRISB7_bit;
// End LCD module connections

//SHT11 connections
sbit SDA1 at RB0_bit;                 // Serial data pin
sbit SCL1 at RB1_bit;                 // Serial clock pin
sbit SDA1_Direction at TRISB0_bit;    // Serial data direction pin
sbit SCL1_Direction at TRISB1_bit;    // Serial clock direction pin
//SHT75 connections
sbit SDA2 at RB2_bit;                 // Serial data pin
sbit SCL2 at RB3_bit;                 // Serial clock pin
sbit SDA2_Direction at TRISB2_bit;    // Serial data direction pin
sbit SCL2_Direction at TRISB3_bit;    // Serial clock direction pin

// Constants for calculating temperature and humidity
// SHT1x/SHT7x V4 humidity conversion coefficients  (12 bits)
const unsigned int C1 = 205;             // -2.0468
const unsigned int C2 = 367;             // 0.0367  (367 * 10^-4)
const unsigned short C3 = 16;           // -1.5955* 10^-6  (15.955 * 10^-7)

// SHT1x/SHT7x V4 temperature compensation coefficients (12 bits)
const unsigned int T1 = 1000;            // 0.01 (1*10^-2)
const unsigned int T2 = 8;               // 0.00008  (8 * 10^-5)

// SHT1x V4 temperature conversion coefficients (14 bits)
const unsigned int D1 = 4010;            // -40.1
const unsigned short D2 = 1;             // 0.01

unsigned short i, j, Select;
long int temp, k, SOt, SOrh, Tmp_res, Rh_res;
char *Tmp = "000.0  ";
char *Rh = "00.0 ";

void SHT_Reset() {
  if(Select == 1) {
  SCL1 = 0;                              // SCL low
  SDA1_Direction = 1;                    // define SDA as input
  for (i = 1; i <= 10; i++)              // repeat 10 times
    SCL1 = ~SCL1;                        // invert SCL
  }
  if(Select == 2) {
  SCL2 = 0;                              // SCL low
  SDA2_Direction = 1;                    // define SDA as input
  for (i = 1; i <= 10; i++)              // repeat 10 times
    SCL2 = ~SCL2;                        // invert SCL
  }
  }


void Transmission_Start() {
  if(Select == 1){
  SDA1_Direction = 1;                     // define SDA as input
  SCL1 = 1;                               // SCL high
  Delay_1us();                           // 1us delay
  SDA1_Direction = 0;                     // define SDA as output
  SDA1 = 0;                               // SDA low
  Delay_1us();                           // 1us delay
  SCL1 = 0;                               // SCL low
  Delay_1us();                           // 1us delay
  SCL1 = 1;                               // SCL high
  Delay_1us();                           // 1us delay
  SDA1_Direction = 1;                     // define SDA as input
  Delay_1us();                           // 1us delay
  SCL1 = 0;                               // SCL low
  }

  if(Select == 2){
  SDA2_Direction = 1;                     // define SDA as input
  SCL2 = 1;                               // SCL high
  Delay_1us();                           // 1us delay
  SDA2_Direction = 0;                     // define SDA as output
  SDA2 = 0;                               // SDA low
  Delay_1us();                           // 1us delay
  SCL2 = 0;                               // SCL low
  Delay_1us();                           // 1us delay
  SCL2 = 1;                               // SCL high
  Delay_1us();                           // 1us delay
  SDA2_Direction = 1;                     // define SDA as input
  Delay_1us();                           // 1us delay
  SCL2 = 0;                               // SCL low
  }
}

// MCU ACK
void MCU_ACK() {

  if(Select == 1){
  SDA1_Direction = 0;     // define SDA as output
  SDA1 = 0;               // SDA low
  SCL1 = 1;               // SCL high
  Delay_1us();           // 1us delay
  SCL1 = 0;               // SCL low
  Delay_1us();           // 1us delay
  SDA1_Direction = 1;     // define SDA as input
  }
  if(Select == 2){
  SDA2_Direction = 0;     // define SDA as output
  SDA2 = 0;               // SDA low
  SCL2 = 1;               // SCL high
  Delay_1us();           // 1us delay
  SCL2 = 0;               // SCL low
  Delay_1us();           // 1us delay
  SDA2_Direction = 1;     // define SDA as input
  }
}

// This function returns temperature or humidity, depends on command
long int Measure(short command) {
  j = command;                           // j = command (0x03 or 0x05)
  SHT_Reset();                       // procedure for reseting SHT11
  Transmission_Start();              // procedure for sTmprting transmission
  k = 0;                             // k = 0
  if(Select == 1){
   SDA1_Direction = 0;                 // define SDA as output
   SCL1 = 0;                           // SCL low
   for(i = 1; i <= 8; i++) {          // repeat 8 times
    if (j.F7 == 1)                   // if bit 7 = 1
     SDA1_Direction = 1;              // define SDA as input
    else {                           // else (if bit 7 = 0)
     SDA1_Direction = 0;              // define SDA as output
     SDA1 = 0;                        // SDA low
   }
    Delay_1us();                     // 1us delay
    SCL1 = 1;                         // SCL high
    Delay_1us();                     // 1us delay
    SCL1 = 0;                         // SCL low
    j <<= 1;                         // move contents of j one place left
   }

   SDA1_Direction = 1;                 // define SDA as input
   SCL1 = 1;                           // SCL high
   Delay_1us();                       // 1us delay
   SCL1 = 0;                           // SCL low
   Delay_1us();                       // 1us delay
   while (SDA1 == 1)                   // while SDA is high, do nothing
    Delay_1us();                     // 1us delay

  for (i = 1; i <=16; i++) {         // repeat 16 times
    k <<= 1;                         // move contents of k one place left
    SCL1 = 1;                         // SCL high
    if (SDA1 == 1)                    // if SDA is high
    k = k | 0x0001;
    SCL1 = 0;
    if (i == 8)                      // if counter i = 8 then
      MCU_ACK();                     // MCU acknowledge
   }
  }

    if(Select == 2){
   SDA2_Direction = 0;                 // define SDA as output
   SCL2 = 0;                           // SCL low
   for(i = 1; i <= 8; i++) {          // repeat 8 times
    if (j.F7 == 1)                   // if bit 7 = 1
     SDA2_Direction = 1;              // define SDA as input
    else {                           // else (if bit 7 = 0)
     SDA2_Direction = 0;              // define SDA as output
     SDA2 = 0;                        // SDA low
   }
    Delay_1us();                     // 1us delay
    SCL2 = 1;                         // SCL high
    Delay_1us();                     // 1us delay
    SCL2 = 0;                         // SCL low
    j <<= 1;                         // move contents of j one place left
   }

   SDA2_Direction = 1;                 // define SDA as input
   SCL2 = 1;                           // SCL high
   Delay_1us();                       // 1us delay
   SCL2 = 0;                           // SCL low
   Delay_1us();                       // 1us delay
   while (SDA2 == 1)                   // while SDA is high, do nothing
    Delay_1us();                     // 1us delay

  for (i = 1; i <=16; i++) {         // repeat 16 times
    k <<= 1;                         // move contents of k one place left
    SCL2 = 1;                         // SCL high
    if (SDA2 == 1)                    // if SDA is high
    k = k | 0x0001;
    SCL2 = 0;
    if (i == 8)                      // if counter i = 8 then
      MCU_ACK();                     // MCU acknowledge
   }
  }

  return k;                          // returns contents of k
}

void STartUpDelay(){
 Delay_ms(20);
}

void DisplayMeasurements(unsigned short rownumber){
 // Calculating temperature
// Tmp_res = D1 + D2 * SOt
    if(SOt > D1)                     // if temperature is positive
      Tmp_res = SOt * D2 - D1;        // calculate temperature
    else                             // else (if temperature is negative)
      Tmp_res = D1 - SOt * D2;        // calculate temperature

// Calculating humidity
// Rh_res = C1 + C2 * SOrh + C3 * SOrh^2
    temp = SOrh * SOrh * C3 / 100000;             // calculate humidity
    Rh_res = SOrh * C2 / 100 - temp - C1;         // calculate humidity
// Temperature compensation
// RHtrue=(T-25)*(T1+T2*SOrh)+RHlin
    temp=(T1+(T2*SOrh))/1000;
    temp=((Tmp_res-2500)*temp)/100;
    Rh_res=temp+Rh_res;

// Preparing temperature for LCD
    Tmp[0] = Tmp_res / 10000 + 48;                  // example: Tmp[5] = 12345 / 10000 = 1, 1 + 48 = '1' - ASCII
    Tmp[1] = Tmp_res % 10000 / 1000 + 48;           // example: Tmp[6] = 12345 % 10000 = 2345, Tmp[6] = 2345 / 1000 = 2, 2 + 48 = '2' - ASCII
    Tmp[2] = Tmp_res % 1000 / 100 + 48;             // example: Tmp[7] = 12345 % 1000 = 345, Tmp[7] = 345 / 100 = 3, 3 + 48 = '3' - ASCII
    Tmp[4] = Tmp_res % 100 / 10 + 48;               // example: Tmp[9] = 12345 % 100 = 45, Tmp[9] = 45 / 10 = 4, 4 + 48 = '4' - ASCII


// Preparing humidity for LCD
    Rh[0] = Rh_res % 10000 / 1000 + 48;           // example: Rh[6] = 12345 % 10000 = 2345, Rh[6] = 2345 / 1000 = 2, 2 + 48 = '2' - ASCII
    Rh[1] = Rh_res % 1000 / 100 + 48;             // example: Rh[7] = 12345 % 1000 = 345, Rh[7] = 345 / 100 = 3, 3 + 48 = '3' - ASCII
    Rh[3] = Rh_res % 100 / 10 + 48;               // example: Rh[9] = 12345 % 100 = 45, Rh[9] = 45 / 10 = 4, 4 + 48 = '4' - ASCII

// delete unnecessary digits (zeros)
    if (Tmp[0] == '0')                             // if Tmp[5] = '0' then
      Tmp[0] = ' ';                                // insert blank character to Tmp[5]
    if (Tmp[0] == ' ' && Tmp[1] == '0')             // if Tmp[5] is blank and Tmp[6] = '0' then
      Tmp[1] = ' ';                                // insert blank character to Tmp[6]

    if (Rh[0] == '0')                             // if Tmp[5] = '0' then
      Rh[0] = ' ';                                // insert blank character to Tmp[5]
    if (Rh[0] == ' ' && Rh[1] == '0')             // if Tmp[5] is blank and Tmp[6] = '0' then
      Rh[1] = ' ';                                // insert blank character to Tmp[6]

// Display temperature on LCD
    Lcd_Out(rownumber, 4, Tmp);                            // display temperature on first row, i column

// Display humidity on LCD
    Lcd_Out(rownumber, 12, Rh);                            // display humidity on second row, i column

}

void main()
{
  CMCON = 0x07;   // Disable comparators
  ADCON1 = 0x0F;  // Disable Analog functions
  TRISC = 0x00;
  TRISB = 0x00;
  TRISC = 0;
  SCL1_Direction = 0;                 // SCL is output
  LCD_Init();                        // init LCD on PORTB (for EasyPIC3 or lower)
  LCD_Cmd(_LCD_CURSOR_OFF);          // turns of LCD cursor
  LCD_Cmd(_LCD_CLEAR);               // clear LCD
  Lcd_Out(1,1, "1:");
  Lcd_Out(2,1, "2:");
  Tmp[5] = 223;                      // If you see greek alpha letter try typing 178 instead of 223
  Tmp[6] = 'C';                      // 'C' character
  Rh[4] = '%';                      // '%' character

  while (1) {
     // Measuring temperature with SHT11
    Select = 1;
    STartUpDelay();
    SCL1_Direction = 0;               // define SCL1 as output
    SOt = Measure(0x03);             // function for measuring (command 0x03 is for temperature)
     // Measuring humidity with SHT11
    SOrh = Measure(0x05);            // function for measuring (command 0x05 is for humidity)
    SCL1_Direction = 1;               // define SCL1 as input
    DisplayMeasurements(1);

     // Measuring temperature with SHT11
    Select = 2;
    STartUpDelay();
    SCL2_Direction = 0;               // define SCL2 as output
    SOt = Measure(0x03);             // function for measuring (command 0x03 is for temperature)
     // Measuring humidity with SHT11
    SOrh = Measure(0x05);            // function for measuring (command 0x05 is for humidity)
    SCL2_Direction = 1;               // define SCL2 as input
    DisplayMeasurements(2);
    Delay_ms(2000);                  // delay 1s
  }
}